# Copyright 2023 Cristian Grosu
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from PIL import Image
import numpy as np
from matplotlib.offsetbox import OffsetImage, AnnotationBbox, TextArea

from decision_boundary_mapper.Logger import Logger

COLORS_MAPPER = {
    -2: [0,0,0], # setting original test data to black
    -1: [1,1,1], # setting original train data to white
    0: [1,0,0], 
    1: [0,1,0], 
    2: [0,0,1], 
    3: [1,1,0], 
    4: [0,1,1], 
    5: [1,0,1], 
    6: [0.5,0.5,0.5], 
    7: [0.5,0,0], 
    8: [0,0.5,0], 
    9: [0,0,0.5]
}

class DBMPlotter:
    def __init__ (self, img, img_confidence, img_projection_errors, img_inverse_projection_errors, 
                  num_classes, encoded_train, encoded_test, X_train, Y_train, X_test, Y_test, logger=None):
        if logger is None:
            self.console = Logger(name="DBMPlotter")
        else:
            self.console = logger
        
        self.img = img
        self.img_confidence = img_confidence
        self.img_projection_errors = img_projection_errors
        self.img_inverse_projection_errors = img_inverse_projection_errors
        self.num_classes = num_classes
        self.encoded_train = encoded_train
        self.encoded_test = encoded_test
        self.X_train = X_train
        self.Y_train = Y_train
        self.X_test = X_test
        self.Y_test = Y_test
        self.color_img, self.legend = self._build_2D_image(img, img_confidence)
        self.train_mapper, self.test_mapper = self._generate_encoded_mapping()
        self._initialize_plot()
    
    def _initialize_plot(self):    
        plt.close("all")
        
        self.fig = plt.figure(figsize = (15, 10))
        self.ax = plt.subplot(1,2,1)
        self.ax_proj_errs = plt.subplot(2, 2, 2) 
        self.ax_inv_proj_errors = plt.subplot(2, 2, 4)
        
        self.ax.xaxis.set_visible(False)
        self.ax.yaxis.set_visible(False)
        self.ax_proj_errs.xaxis.set_visible(False)
        self.ax_proj_errs.yaxis.set_visible(False)
        self.ax_inv_proj_errors.xaxis.set_visible(False)
        self.ax_inv_proj_errors.yaxis.set_visible(False)
        self._build_annotation_mapper()
    
    def _build_2D_image(self, img, img_confidence, class_name_mapper = lambda x: str(x), colors_mapper = COLORS_MAPPER):
        color_img = np.zeros((img.shape[0], img.shape[1], 4))
                
        for i, j in np.ndindex(img.shape):
            color_img[i,j] = colors_mapper[img[i,j]] + [img_confidence[i,j]]
        
        values = np.unique(img)
        
        patches = []
        for value in values:
            color = colors_mapper[value]
            if value==-1:
                label = "Original train data"
            elif value==-2:
                label = "Original test data"
            else:
                label = f"Value region: {class_name_mapper(value)}"

            patches.append(mpatches.Patch(color=color, label=label))
        
        return color_img, patches
    
    def _generate_encoded_mapping(self):
        train_mapper = {}
        for k in range(len(self.encoded_train)):
            [x, y] = self.encoded_train[k]
            train_mapper[f"{x} {y}"] = k
    
        test_mapper = {}
        for k in range(len(self.encoded_test)):
            [x, y] = self.encoded_test[k]
            test_mapper[f"{x} {y}"] = k
        
        return train_mapper, test_mapper

    def _build_annotation_mapper(self):
        image = OffsetImage(self.X_train[0], zoom=2, cmap="gray")
        label = TextArea("Data point label: None")

        xybox=(50., 50.)
        annImage = AnnotationBbox(image, (0,0), xybox=xybox, xycoords='data',
                boxcoords="offset points",  pad=0.1,  arrowprops=dict(arrowstyle="->"))
        annImage_ax1 = AnnotationBbox(image, (0,0), xybox=xybox, xycoords='data',
                boxcoords="offset points",  pad=0.1,  arrowprops=dict(arrowstyle="->"))
        annImage_ax2 = AnnotationBbox(image, (0,0), xybox=xybox, xycoords='data',
                boxcoords="offset points",  pad=0.1,  arrowprops=dict(arrowstyle="->"))

        annLabels = AnnotationBbox(label, (0,0), xybox=xybox, xycoords='data',
                boxcoords="offset points",  pad=0.3,  arrowprops=dict(arrowstyle="->"))

        self.ax.add_artist(annImage)
        self.ax.add_artist(annLabels)
        self.ax_inv_proj_errors.add_artist(annImage_ax1)
        self.ax_proj_errs.add_artist(annImage_ax2)
        annImage.set_visible(False)
        annImage_ax1.set_visible(False)
        annImage_ax2.set_visible(False)
        annLabels.set_visible(False)

        fig_width, fig_height = self.fig.get_size_inches() * self.fig.dpi
            
        def display_annotation(event):
            if event.inaxes == None:
                return
        
            j, i = int(event.xdata), int(event.ydata)
            if self.img[i,j] >= 0 or self.img[i,j] < -2:                
                return
            
            # change the annotation box position relative to mouse.
            ws = (event.x > fig_width/2.)*-1 + (event.x <= fig_width/2.) 
            hs = (event.y > fig_height/2.)*-1 + (event.y <= fig_height/2.)
            annImage.xybox = (50. * ws, 50. * hs)
            annLabels.xybox = (-50. * ws, 50. *hs)
            # make annotation box visible
            annImage.set_visible(True)
            annImage_ax1.set_visible(True)
            annImage_ax2.set_visible(True)
            annLabels.set_visible(True)
            # place it at the position of the event scatter point
            annImage.xy = (j, i)
            annImage_ax1.xy = (j, i)
            annImage_ax2.xy = (j, i)
            annLabels.xy = (j, i)

            x_data, y_data = find_data_point(i,j)                            
            image.set_data(x_data)
            label.set_text(f"Label: {y_data}")

            self.fig.canvas.draw_idle()
                
        def onclick(event):
            if event.inaxes == None:
                return
            
            self.console.log("Clicked on: " + str(event.xdata) + ", " + str(event.ydata))
            j, i = int(event.xdata), int(event.ydata)

            if self.img[i,j] >= 0 or self.img[i,j] < -2:
                self.console.log("Data point not in training or testing set")

            x, y = find_data_point(i, j)
            self.plot_data_point(x, y)
            
            
        def find_data_point(i, j):
            if self.img[i,j] == -1:
                k = self.train_mapper[f"{i} {j}"]
                self.console.log("Found point in train set")
                return self.X_train[k], self.Y_train[k]
            
            if self.img[i,j] == -2:
                k = self.test_mapper[f"{i} {j}"]
                self.console.log("Found point in test set")
                return self.X_test[k], self.Y_test[k]        
            
            return None, None
            
        self.fig.canvas.mpl_connect('motion_notify_event', display_annotation)           
        self.fig.canvas.mpl_connect('button_press_event', onclick)
    
    def plot_data_point(self, data, label):
        img = Image.fromarray(np.uint8(data * 255))
        img.thumbnail((100, 100), Image.ANTIALIAS)
        img.show(title=f"Data point label: {label}")
        
    def plot(self, title="Decision Boundary Mapper"):
        self._initialize_plot()
        self.ax.imshow(self.color_img)
        self.ax.set_title(title)
        self.ax.legend(handles=self.legend, bbox_to_anchor=(0, 1), loc=1, borderaxespad=0. )
        
        ax_img1 = self.ax_proj_errs.imshow(self.img_projection_errors, cmap="Reds")
        self.fig.colorbar(ax_img1, ax=self.ax_proj_errs)
        self.ax_proj_errs.set_title("Projection Errors")
        
        ax_img2 = self.ax_inv_proj_errors.imshow(self.img_inverse_projection_errors, cmap="Reds")
        self.fig.colorbar(ax_img2, ax=self.ax_inv_proj_errors)
        self.ax_inv_proj_errors.set_title("Inverse Projection Errors")
        
        self.fig.show()
        plt.show()


# Copyright 2023 Cristian Grosu
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tensorflow as tf
import matplotlib.pyplot as plt
import numpy as np

from decision_boundary_mapper.Logger import Logger, LoggerInterface

DEFAULT_MODEL_PATH = os.path.join("models", "SDBM")

def build_autoencoder(classifier, input_shape:tuple = (28, 28), show_summary:bool = False):
    """ Building an autoencoder for dimensionality reduction of the data

    Args:
        classifier (tensorflow.keras.Model): The classifier to be used for the autoencoder
        input_shape (tuple, optional): Data input shape. Defaults to (28, 28).
        num_classes (int, optional): Target number of classes to be predicted. Defaults to 10.
        show_summary (bool, optional): If true shows the autoencoder summary. Defaults to False.

    Returns:
        autoencoder (tensorflow.keras.Model): The autoencoder model
    """
    ENCODER = tf.keras.models.Sequential([
        tf.keras.layers.Flatten(),
        tf.keras.layers.Dense(512, activation='relu', bias_initializer=tf.keras.initializers.Constant(10e-4)),
        tf.keras.layers.Dense(128, activation='relu', bias_initializer=tf.keras.initializers.Constant(10e-4)),
        tf.keras.layers.Dense(32, activation='relu', bias_initializer=tf.keras.initializers.Constant(10e-4)),
        tf.keras.layers.Dense(2, activation='linear', activity_regularizer=tf.keras.regularizers.l2(1/2), bias_initializer=tf.keras.initializers.Constant(10e-4)),
    ], name="encoder")

    DECODER = tf.keras.models.Sequential([
        tf.keras.layers.Dense(32, activation='relu', bias_initializer=tf.keras.initializers.Constant(10e-4)),
        tf.keras.layers.Dense(128, activation='relu', bias_initializer=tf.keras.initializers.Constant(10e-4)),
        tf.keras.layers.Dense(512, activation='relu', bias_initializer=tf.keras.initializers.Constant(10e-4)),
        tf.keras.layers.Dense(input_shape[0] * input_shape[1], activation='sigmoid'),
        tf.keras.layers.Reshape(input_shape)
    ], name="decoder")       
    
    
    INPUT = tf.keras.Input(shape=input_shape, name="input")
    
    autoencoder = Autoencoder(ENCODER, DECODER, classifier, INPUT)
    
    if show_summary:
        autoencoder.summary()
        
    return autoencoder

class Autoencoder:
    def __init__(self, encoder = None, decoder = None, classifier = None, input_layer = None, 
                 folder_path:str = DEFAULT_MODEL_PATH, load:bool = False, logger:LoggerInterface = None):
        """
            Creates an autoencoder model.
            Encoder: The encoder part of the autoencoder.
            Decoder: The decoder part of the autoencoder.
            Classifier: The classifier part of the autoencoder.
            Input_layer: The input layer of the autoencoder.
        """
        if logger is not None:
            self.console = logger
        else:
            self.console = Logger(name="Autoencoder")
        
        self.save_folder_path = folder_path
        self.encoder = encoder
        self.decoder = decoder
        self.classifier = classifier
        self.input_layer = input_layer
        
        if load:
            self.load(folder_path)
        else:
            self.build()
    
    def load(self, folder_path:str):
        """
            Loads an autoencoder from the specified folder path. With the .h5 extension.
            Args:
                folder_path (str): The folder path where the autoencoder is saved.
        """
        self.save_folder_path = folder_path
        try:
            self.autoencoder_classifier = tf.keras.models.load_model(os.path.join(folder_path, "autoencoder_classifier.h5"), compile=False)
            self.auto_encoder = tf.keras.models.load_model(os.path.join(folder_path, "auto_encoder.h5"), compile=False)
            self.classifier = tf.keras.models.load_model(os.path.join(folder_path, "classifier.h5"), compile=False)
            self.autoencoder = tf.keras.models.load_model(os.path.join(folder_path, "autoencoder.h5"), compile=False)
            self.decoder = tf.keras.models.load_model(os.path.join(folder_path, "decoder.h5"), compile=False)
            self.encoder = tf.keras.models.load_model(os.path.join(folder_path, "encoder.h5"), compile=False)
        except Exception as e:
            self.console.log("Autoencoder not found. Please check the path folder and make sure the autoencoder is saved there")        
            self.console.error(f"Exception: {e}")
            raise e
    
    
    def build(self):
        """Assembles the autoencoder model and compiles it. 
        """
        self.autoencoder_classifier = tf.keras.models.Sequential([
            self.encoder,
            self.decoder,
            self.classifier,
        ], name="autoencoder_classifier")
    
        self.auto_encoder = tf.keras.models.Sequential([
            self.encoder,
            self.decoder,
        ], name="auto_encoder")
        
        auto_encoder_classifier_output = self.autoencoder_classifier(self.input_layer)
        auto_encoder_output = self.auto_encoder(self.input_layer)
        
            
        self.autoencoder = tf.keras.models.Model(inputs=self.input_layer, 
                                            outputs=[auto_encoder_output, 
                                                    auto_encoder_classifier_output], 
                                            name="autoencoder")
        
        optimizer = tf.keras.optimizers.Adam()
        
        self.autoencoder.compile(optimizer=optimizer, 
                                loss={"auto_encoder":"binary_crossentropy",
                                      "autoencoder_classifier": "sparse_categorical_crossentropy"}, 
                                metrics=['accuracy'])
    
    def summary(self):
        self.autoencoder.summary()
        
    def fit(self, x_train: np.ndarray, y_train: np.ndarray, 
            x_test: np.ndarray, y_test: np.ndarray,
            epochs:int = 10, batch_size:int = 128):
        """ Fits the model to the specified data.

        Args:
            x_train (np.ndarray): Train input values
            y_train (np.ndarray): Train target values
            x_test (np.ndarray): Test input values
            y_test (np.ndarray): Test target values
            epochs (int, optional): The number of epochs. Defaults to 10.
            batch_size (int, optional): Data points used for one batch. Defaults to 128.
        """
        self.autoencoder.fit(x_train, [x_train, y_train], 
                            epochs=epochs, 
                            batch_size=batch_size, 
                            shuffle=True, 
                            validation_data=(x_test, [x_test, y_test]))
        
        self.save()
        
    def save(self):
        """
            Saves the model to the specified folder path. With the .h5 extension.
        """       
        folder_path = self.save_folder_path 
        if not os.path.exists(folder_path):
            os.makedirs(folder_path)
            
        self.auto_encoder.save(os.path.join(folder_path, "auto_encoder.h5"))
        self.autoencoder_classifier.save(os.path.join(folder_path, "autoencoder_classifier.h5"))
        self.autoencoder.save(os.path.join(folder_path, "autoencoder.h5"))
        self.decoder.save(os.path.join(folder_path, "decoder.h5"))
        self.encoder.save(os.path.join(folder_path, "encoder.h5"))
        self.classifier.save(os.path.join(folder_path, "classifier.h5"))
    
    def show_predictions(self, data:np.ndarray, labels:np.ndarray):
        """Shows the predictions of the autoencoder. First 20 data points from the provided data input.

        Args:
            data (np.ndarray): The data set to be predicted.
            labels (np.ndarray): The actual labels of the data set.
        """
        decoded = self.auto_encoder.predict(data)
        decoded_labels = self.autoencoder_classifier.predict(data)
        predicted_labels = [np.argmax(label) for label in decoded_labels]
        
        plt.figure(figsize=(20, 6))
        
        for i in range(1,20,2):
            plt.subplot(2, 10, i)
            plt.imshow(data[i], cmap='gray')
            plt.axis('off')
            plt.subplot(2, 10, i+1)
            plt.imshow(decoded[i], cmap='gray')
            plt.title(f"Predicted: {predicted_labels[i]} \n Actual: {labels[i]}", color='green' if predicted_labels[i] == labels[i] else 'red')
            plt.axis('off')
        
        plt.show()
    
    def encode(self, data:np.ndarray):
        #self.console.log("Encoding data")
        return self.encoder.predict(data)
    
    def decode(self, data:np.ndarray):
        #self.console.log("Decoding data")
        return self.decoder.predict(data, verbose=0)
    


# Copyright 2023 Cristian Grosu
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tensorflow as tf
import numpy as np
import matplotlib.pyplot as plt

from decision_boundary_mapper.Logger import Logger, LoggerInterface

DEFAULT_MODEL_PATH = os.path.join("models", "DBM")

class invNN:
    """
        Inverse Projection Neural Network.
        The network is composed of 2 parts:
            1) The decoder part, i.e. The inverse projection 2D -> nD. (Sequential model 2 -> 32 -> 128 -> 512 -> nD) if is not provided by the user
            2) The classifier part nD -> 1D (provided by the user).
    """
    
    def __init__(self, 
                 decoder = None, 
                 classifier = None, 
                 logger: LoggerInterface = None, 
                 load: bool = False,
                 folder_path: str = DEFAULT_MODEL_PATH):
        """
            Creates an inverse Projection Neural Network model.
            Decoder: The decoder part, i.e. The inverse projection 2D -> nD.
            Classifier: The classifier part nD -> 1D.
        """
        if logger is not None:
            self.console = logger
        else:
            self.console = Logger(name="inverse Projection Neural Network")
        
        self.save_folder_path = folder_path
        self.decoder = decoder
        self.classifier = classifier
        
        if load:
            self.load(folder_path)
            
        self.pre_load = load
        
    def load(self, folder_path:str):
        """
            Loads an auto encoder from the specified folder path. With the .h5 extension.
            Args:
                folder_path (str): The path to the folder where the model is saved.
        """
        self.save_folder_path = folder_path
        try:
            self.classifier = tf.keras.models.load_model(os.path.join(folder_path, "classifier.h5"), compile=False)
            self.decoder = tf.keras.models.load_model(os.path.join(folder_path, "decoder.h5"), compile=False)
            self.decoder_classifier = tf.keras.models.load_model(os.path.join(folder_path, "decoder_classifier.h5"), compile=False)
            self.invNN = tf.keras.models.load_model(os.path.join(folder_path, "invNN.h5"), compile=False)
        except Exception as e:
            self.console.log(f"Inverse projection NN not found. Please check the path folder {folder_path} and make sure the model is saved there")        
            self.console.error(f"Exception: {e}")
            raise e
    
    def __build__(self, output_shape:tuple = (2,1)):
        """Builds an invNN if not provided by the user (Sequential 2D -> 32 -> 128 -> 512 -> nD -> ... -> 1D)

        Args:
            output_shape (tuple, optional): _description_. Defaults to (2,1).
        """
        
        assert len(output_shape) == 2, "Output shape must be a 2D tuple"
        
        self.decoder = tf.keras.Sequential([
            tf.keras.layers.Dense(32, activation='relu'),
            tf.keras.layers.Dense(128, activation='relu'),
            tf.keras.layers.Dense(512, activation='relu'),
            tf.keras.layers.Dense(output_shape[0] * output_shape[1], activation='linear'),
            tf.keras.layers.Reshape(output_shape)
        ], name="decoder")

        self.decoder_classifier = tf.keras.Sequential([
            self.decoder,
            self.classifier
        ], name="decoder_classifier")  
        
        input_layer = tf.keras.Input(shape=(2,), name="input")
                                             
        decoder_classifier_output = self.decoder_classifier(input_layer)
        decoder_output = self.decoder(input_layer)
            
        self.invNN = tf.keras.models.Model(inputs=input_layer, 
                                            outputs=[decoder_output, 
                                                    decoder_classifier_output], 
                                            name="invNN")
        
        optimizer = tf.keras.optimizers.Adam()
        
        self.invNN.compile(optimizer=optimizer, 
                                loss={"decoder":"binary_crossentropy",
                                      "decoder_classifier": "sparse_categorical_crossentropy"}, 
                                metrics=['accuracy'])
                                             
    def summary(self):
        self.invNN.summary()
        
    def fit(self, 
            x2d_train: np.ndarray, xNd_train: np.ndarray, y_train: np.ndarray, 
            x2d_test: np.ndarray, xNd_test: np.ndarray, y_test: np.ndarray, 
            epochs:int = 10, batch_size:int = 128):
        """ Fits the model to the specified data.

        Args:
            x2d_train (np.ndarray): Train input values (2D)
            xNd_train (np.ndarray): Train input values (nD)
            y_train (np.ndarray): Train target values
            x2d_test (np.ndarray): Test input values (2D)
            xNd_test (np.ndarray): Test input values (nD)
            y_test (np.ndarray): Test target values
            epochs (int, optional): The number of epochs. Defaults to 10.
            batch_size (int, optional): Data points used for one batch. Defaults to 128.
        """
        if self.pre_load:
            self.console.log("Model already loaded. Skipping build.")
        else:
            self.console.log("Building model according to the data shape.")
            self.__build__(xNd_train.shape[1:])
            
        self.console.log("Fitting model...")
        self.invNN.fit(x2d_train, [xNd_train, y_train], 
                            epochs=epochs, 
                            batch_size=batch_size, 
                            shuffle=True, 
                            validation_data=(x2d_test, [xNd_test, y_test]))
        self.console.log("Model fitted!")
        self.save()
        
    def save(self):
        """
            Saves the model to the specified folder path. With the .h5 extension.
        """       
        folder_path = self.save_folder_path 
        if not os.path.exists(folder_path):
            os.makedirs(folder_path)
            
        self.classifier.save(os.path.join(folder_path, "classifier.h5"))
        self.decoder.save(os.path.join(folder_path, "decoder.h5"))
        self.decoder_classifier.save(os.path.join(folder_path, "decoder_classifier.h5"))
        self.invNN.save(os.path.join(folder_path, "invNN.h5"))
        self.console.log(f"Model saved to {folder_path}")
        
    def show_predictions(self, data: np.ndarray, labels: np.ndarray):
        """ Shows the predictions of the model. By taking first 20 data points and comparing them to the actual labels.

        Args:
            data (np.ndarray): Data set
            labels (np.ndarray): Actual labels of the data points
        """
        decoded = self.decoder.predict(data)
        decoded_labels = self.decoder_classifier.predict(data)
        predicted_labels = [np.argmax(label) for label in decoded_labels]
        
        plt.figure(figsize=(20, 6))
        
        for i in range(1,20,2):
            plt.subplot(2, 10, i)
            plt.imshow(data[i], cmap='gray')
            plt.axis('off')
            plt.subplot(2, 10, i+1)
            plt.imshow(decoded[i], cmap='gray')
            plt.title(f"Predicted: {predicted_labels[i]} \n Actual: {labels[i]}", color='green' if predicted_labels[i] == labels[i] else 'red')
            plt.axis('off')
        
        plt.show()
    
    def decode(self, data: np.ndarray):
        return self.decoder.predict(data, verbose=0)
    


# Copyright 2023 Cristian Grosu
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time

def track_time_wrapper(func):
    """ Decorator that tracks the time it takes for a function to run.

    Args:
        func (python function): The function to be decorated.
    """
    def wrapper(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        print(f"{func.__name__} took {end-start} seconds")
        return result
    return wrapper
